#include <stdio.h>
#include <assert.h>
#include "opengl.h"
#include "app.h"
#include "gmath/gmath.h"
#include "mesh.h"
#include "meshgen.h"
#include "backdrop.h"
#include "goatvr.h"

static bool parse_args(int argc, char **argv);

int win_width, win_height;
float win_aspect;
long time_msec;
Mat4 view_matrix;

static bool use_vr;
static bool should_swap;

static float cam_theta, cam_phi;
static float cam_height = 1.65;
static Mesh *mesh_torus;

static bool bnstate[16];
static int prev_x, prev_y;

bool app_init(int argc, char **argv)
{
	if(!parse_args(argc, argv)) {
		return false;
	}
	if(init_opengl() == -1) {
		return false;
	}

	glEnable(GL_MULTISAMPLE);

	int aasamples = 0;
	glGetIntegerv(GL_SAMPLES, &aasamples);
	printf("got %d samples per pixel\n", aasamples);

	printf("Max anisotropy: %d\n", glcaps.max_aniso);

	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	//glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	if(GLEW_ARB_framebuffer_sRGB) {
		glEnable(GL_FRAMEBUFFER_SRGB);
	}

	if(use_vr) {
		if(goatvr_init() == -1) {
			return false;
		}
		goatvr_set_origin_mode(GOATVR_HEAD);

		goatvr_startvr();
		should_swap = goatvr_should_swap() != 0;
		cam_height = goatvr_get_eye_height();
	}

	Mesh::use_custom_sdr_attr = false;

	mesh_torus = new Mesh;
	gen_torus(mesh_torus, 1.0, 0.25, 32, 32);

	if(!init_backdrop()) {
		return false;
	}

	return true;
}

void app_cleanup()
{
	if(use_vr) {
		goatvr_shutdown();
	}
	delete mesh_torus;
	cleanup_backdrop();
}

void app_draw()
{
	if(use_vr) {
		// VR mode
		goatvr_draw_start();
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		for(int i=0; i<2; i++) {
			goatvr_draw_eye(i);

			glMatrixMode(GL_PROJECTION);
			glLoadMatrixf(goatvr_projection_matrix(i, 0.5, 1000.0));

			view_matrix = goatvr_view_matrix(i);
			view_matrix.pre_rotate_x(deg_to_rad(cam_phi));
			view_matrix.pre_rotate_y(deg_to_rad(cam_theta));
			view_matrix.pre_translate(0, -cam_height, 0);

			glMatrixMode(GL_MODELVIEW);
			glLoadMatrixf(view_matrix[0]);

			draw_backdrop();
		}
		goatvr_draw_done();

		if(should_swap) {
			app_swap_buffers();
		}
		app_redraw();	// in VR mode, force continuous redraw

	} else {
		// regular monoscopic mode
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		view_matrix = Mat4::identity;
		view_matrix.pre_rotate_x(deg_to_rad(cam_phi));
		view_matrix.pre_rotate_y(deg_to_rad(cam_theta));
		view_matrix.pre_translate(0, -cam_height, 0);

		glMatrixMode(GL_MODELVIEW);
		glLoadMatrixf(view_matrix[0]);

		draw_backdrop();

		app_swap_buffers();
	}
	assert(glGetError() == GL_NO_ERROR);
}

void app_reshape(int x, int y)
{
	glViewport(0, 0, x, y);

	Mat4 mat;
	mat.perspective(deg_to_rad(50), win_aspect, 0.5, 500.0);

	glMatrixMode(GL_PROJECTION);
	glLoadMatrixf(mat[0]);
}

void app_keyboard(int key, bool pressed)
{
	if(pressed) {
		switch(key) {
		case 27:
			app_quit();
			break;

		case ' ':
			if(use_vr) {
				goatvr_recenter();
			}
			break;
		}
	}
}

void app_mouse_button(int bn, bool pressed, int x, int y)
{
	bnstate[bn] = pressed;
	prev_x = x;
	prev_y = y;
}

void app_mouse_motion(int x, int y)
{
	int dx = x - prev_x;
	int dy = y - prev_y;
	prev_x = x;
	prev_y = y;

	if(!dx && !dy) return;

	if(bnstate[0]) {
		cam_theta += dx * 0.5;

		if(!use_vr || !goatvr_have_headtracking()) {
			cam_phi += dy * 0.5;

			if(cam_phi < -90) cam_phi = -90;
			if(cam_phi > 90) cam_phi = 90;
		}
	}
	app_redraw();
}

static bool parse_args(int argc, char **argv)
{
	for(int i=1; i<argc; i++) {
		if(argv[i][0] == '-') {
			if(strcmp(argv[i], "-vr") == 0) {
				use_vr = true;

			} else if(strcmp(argv[i], "-novr") == 0) {
				use_vr = false;

			} else if(strcmp(argv[i], "-help") == 0) {
				printf("usage: %s [options]\noptions:\n", argv[0]);
				printf(" -vr/-novr: enable/disable VR\n");
				printf(" -help: print usage information and exit\n");
				exit(0);
			} else {
				fprintf(stderr, "invalid option: %s\n", argv[i]);
				return false;
			}
		} else {
			fprintf(stderr, "unexpected option: %s\n", argv[i]);
			return false;
		}
	}
	return true;
}
